/**************************************************************************\
 *  aRT : API R - TerraLib                                                *
 *  Copyright (C) 2003-2007  LEG                                          *
 *                                                                        *
 *  This program is free software; you can redistribute it and/or modify  *
 *  it under the terms of the GNU General Public License as published by  *
 *  the Free Software Foundation; either version 2 of the License, or     *
 *  (at your option) any later version.                                   *
 *                                                                        *
 *  This program is distributed in the hope that it will be useful,       *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *  GNU General Public License for more details.                          *
 *                                                                        *
 *  You should have received a copy of the GNU Lesser General Public      *
 *  License along with this library.                                      *
\**************************************************************************/

#include "aRTgpm.h"

//#include <TeProxMatrixConstructionStrategy.h>
#include <TeSTElementSet.h>

/*
extern "C"
{

TeProxMatrixImplementation* GetTeProxMatrixImplementation(SEXP options)
{
	// Until now, TerraLib has only one implementation, therefore there is no need of having this.
}

TeProxMatrixConstructionStrategy* GetTeProxMatrixConstructionStrategySTE(SEXP options, 
																		 TeSTElementSet* objs,
																		 TeGeomRep rep,
																		 TeGraphNetwork* net,
																		 TeSTElementSet* objs2,
																		 TeGeomRep rep2)
{
	string strategy  = GET_STRING_ELEMENT (options, "strategy"    );
	bool calc_dist   = GET_BOOL_ELEMENT   (options, "calcDistance");
	double max_d     = GET_REAL_ELEMENT   (options, "maxDistance" );//2000; // TODO: 2.000.000 da pau!!???
	double max_local_d, max_net_d, max_connect_d; // d = distance

	max_local_d = max_net_d = max_connect_d = max_d;

	switch(strategy[0])
	{
		case 'l': 
		//	if(rep == TePOINTS) return new TeProxMatrixLocalDistanceStrategy<TeSTEventSet> (objs, rep, max_d);
		//	else     
		      return new TeProxMatrixLocalDistanceStrategy<TeSTElementSet>(objs, rep, max_d);
		case 'a': return new TeProxMatrixLocalAdjacencyStrategy (objs, rep, calc_dist);
		case 'c': return new TeProxMatrixClosedNetworkStrategy  (objs, rep,              max_local_d, max_net_d, max_connect_d, net);
		case 'o': return new TeProxMatrixOpenNetworkStrategy    (objs, rep,              max_local_d, max_net_d, max_connect_d, net);
		case 'r': return new TeProxMatrixOpenNetworkStrategy2   (objs, rep, objs2, rep2, max_local_d, max_net_d, max_connect_d, net);
		default: return NULL;
	}
	return NULL;
}

TeProxMatrixProcessingStrategy*  GetTeProxMatrixSlicingStrategy(string s, SEXP options)
{
	double threshold = GET_REAL_ELEMENT(options, "threshold");
	double zones     = GET_REAL_ELEMENT(options, "zones"    );

	switch(s[0])
	{
		case 'i':           return new TeProxMatrixInvertStrategy   ();
		case 'n':
			if(s[1] == 'o') return new TeProxMatrixNormalizeStrategy();
			else            return new TeProxMatrixNearestStrategy  ();
		case 't':           return new TeProxMatrixThresholdStrategy(threshold);
		case 'z':           return new TeProxMatrixZonesStrategy    (zones);	

		default: return NULL;
	}
}

aRTgpm::aRTgpm(SEXP options, TeSTElementSet* eset, TeGeomRep rep, aRTnetwork* net,  vector<string>* ids, TeSTElementSet* targetset, TeGeomRep targetrep)
{
	IDs = ids;

	PrintSilent("Creating strategy ... ");
    TeProxMatrixConstructionStrategy* pmcs;
	if(net) pmcs = GetTeProxMatrixConstructionStrategySTE(options, eset, rep, net->GetNet(), targetset, targetrep );
	else    pmcs = GetTeProxMatrixConstructionStrategySTE(options, eset, rep, NULL,          targetset, targetrep );
	PrintSilentYes;

	PrintSilent("Creating matrix ... ");
	GPM = new TeGeneralizedProxMatrix(pmcs);
	PrintSilentYes;
}

aRTgpm::~aRTgpm()
{

}

SEXP aRTgpm::GetMatrix(SEXP data)
{
	const bool weights = GET_BOOL_ELEMENT(data, "weight");
	const int iw = weights? 1 : 0;

	SEXP neighs, ids, result, names, matweig;

	PROTECT( neighs = allocVector(VECSXP, IDs->size() ) );
	PROTECT( ids    = allocVector(STRSXP, IDs->size() ) ); 

	if(weights)
	{
		PROTECT( matweig = allocMatrix(REALSXP, IDs->size(), IDs->size() ) );

		for(int i = 0; i < IDs->size(); i++)
			for(int j = 0; j < IDs->size(); j++)
				REAL(matweig)[j + IDs->size()*i] = 0.0;
	}

	// CODE DUMP! See why below.
	//	GPM->saveTextFile("matviz.txt", IDs);

	map<string, int> map_;

    vector<string>::iterator it = IDs->begin();
    int Id = 0;
    while( it != IDs->end() )
	{
		map_[(*it)] = Id;
		Id++;
		it++;
	}

    it = IDs->begin();
    Id=0;
    while( it != IDs->end() )
    {
        string objId1, objId2;
        objId1 = (*IDs)[Id];

        TeNeighbours neigs;

        GPM->imp_->getNeighbours((*it), neigs);

		// CORE DUMP!!! don't know why...
		// cout << neigs.size() << endl;

		int count = 0;
        for(TeNeighbours::iterator j=neigs.begin(); j!=neigs.end(); j++, count++);

		SEXP values;

		PROTECT( values = allocVector(INTSXP, count) ); 

		count = 0;
        for(TeNeighbours::iterator j=neigs.begin(); j!=neigs.end(); j++)
        {
			// j->first  => string 
			// j->second => double
			
			INTEGER(values)[count] = map_[j->first] + 1; // positions in R start @1

			if(weights) REAL(matweig)[map_[j->first]*IDs->size() + Id] = j->second;

			count++;
		}
		SET_STRING_ELT(ids,    Id, mkChar((*IDs)[Id].c_str()) );
    	SET_VECTOR_ELT(neighs, Id, values);

        ++it;
		Id++;
	}
	PROTECT( result = allocVector( VECSXP, 2 + iw));
	PROTECT( names  = allocVector( STRSXP, 2 + iw));

	SET_STRING_ELT(names, 0, mkChar("neighbors"));
	SET_STRING_ELT(names, 1, mkChar("ID")       );
	if(weights) SET_STRING_ELT(names, 2, mkChar("weights") );

    setAttrib(result, R_NamesSymbol, names);

    SET_VECTOR_ELT(result, 0, neighs);
    SET_VECTOR_ELT(result, 1, ids);
    if(weights) SET_VECTOR_ELT(result, 2, matweig);

	UNPROTECT( IDs->size() + 4 + iw);
	
	return result;
}

SEXP aRTgpm::Process(SEXP options)
{
	SEXP strats = GetListElement(options, "functions"); 
	int count = LENGTH(strats);

	PrintSilent("Creating operators ... ");
	vector<TeProxMatrixProcessingStrategy*> processors;
	for(int i = 0; i < count; i++)
	{
		string strat = (char*) STRING_PTR( VECTOR_ELT(strats,i) );
		processors.push_back( GetTeProxMatrixSlicingStrategy(strat, options) );
	}
	PrintSilentYes;

	PrintSilent("Processing matrix ... ");
	GPM->Apply (processors);
	PrintSilentYes;
	
	PrintSilent("Removing operators ... ");
	for(int i = 0; i < count; i++)
	{
		delete processors[i];
	}
	PrintSilentYes;

	return RNULL;
}


SEXP aRTgpm::Summary()
{
}

SEXP aRTgpm::Print(){ } 

}
*/


